#ifndef _DEFS_H
#define _DEFS_H 1

#define MAX_STREAMS 16
#define CACHE_SIZE_MB 2048
#define MUX_BUFFER_SIZE 10485760
#define FRAMES_PER_FILE 1000

#define DEFAULT_HTTPD_PORT 9096

#endif  // !defined(_DEFS_H)
