#ifndef _FF_MAYBE_CONST
#define _FF_MAYBE_CONST 1

// Used as a shim between FFmpeg pre- and post-7.0.
#include <libavformat/version_major.h>

#if LIBAVFORMAT_VERSION_MAJOR < 61
#define FF_MAYBE_CONST
#else
#define FF_MAYBE_CONST const
#endif

#endif  // !defined(_FF_MAYBE_CONST)
